@@ -141,7 +141,7 @@ my %allowed_tagnames = (
         NOPARSE     => [undef, qw(NAME)],
         LOOP        => [$default_validation, qw(NAME ALIAS JOIN BREAK EXPR CONTEXT)],
         WHILE       => [$default_validation, qw(NAME ALIAS BREAK EXPR)],
-        EACH        => [$default_validation, qw(NAME BREAK EXPR SORT REVERSE CONTEXT)],
+        EACH        => [$default_validation, qw(NAME BREAK EXPR SORT SORTBY REVERSE CONTEXT)],
         SWITCH      => [$default_validation, qw(NAME EXPR)],
         CASE        => [undef, qw(NAME)],
         INCLUDE_VAR => [$default_validation, qw(NAME EXPR)],
@@ -327,7 +327,7 @@ sub find_attributes {
         ]);
         push @{$arg->{tags}}, $token;
         if ($token->is_open &&
-            not exists
+            exists
                 $self->get_tagnames->{CLOSING_TAG()}->{ $arg->{name} }) {
             $arg->{level}++
         }
